"""Basic infrastructure CLI."""

import functools
import socket
from collections.abc import Collection

import typer
from slices_clientlib_bi.bi_resource import ComputeResource, ResourceState, SSHInfo
from slices_clientlib_core.base.exceptions import ApiResponseError, NotFoundException
from slices_clientlib_core.experiment import Experiment
from urllib3.connection import HTTPSConnection
from urllib3.exceptions import MaxRetryError, NewConnectionError

from slices_cli_bi import err_console
from slices_cli_bi.util.type_id import is_type_id

from .bi_context import BIContext


def bi_get_resource(ctx: BIContext, experiment: Experiment, resource_id: str) -> ComputeResource:
    """Get a resource for a bi command."""
    if is_type_id(resource_id, type="r"):
        try:
            return ctx.obj.get_bi_resources_api_by_experiment(experiment).get(resource_id)
        except NotFoundException as err:
            err_console.print(
                f"[red]Resource {resource_id} not found in experiment {experiment.friendly_name} "
                f"on infrastructure {ctx.obj.infra_id} :x:"
            )
            raise typer.Exit(code=1) from err

    resources = ctx.obj.get_bi_resources_api_by_experiment(experiment).list()

    inactive_resource = None

    def _is_inactive(resource: ComputeResource) -> bool:
        return resource.status in (ResourceState.DELETED, ResourceState.FAILED)

    for resource in resources:
        if resource.friendly_name.lower() == resource_id.lower():
            if _is_inactive(resource):
                if (
                    # no inactive resource with this name found yet
                    not inactive_resource
                    # more recent inactive resource has precedence
                    or resource.created_at > inactive_resource.created_at
                ):
                    inactive_resource = resource
            else:
                return resource

    # no active resource found, return inactive resource if found
    if inactive_resource:
        return inactive_resource

    err_console.print(
        f"[red]Resource {resource_id} not found in experiment {experiment.friendly_name} "
        f"on infrastructure {ctx.obj.infra_id} :x:"
    )
    raise typer.Exit(code=1)


def bi_get_requested_bi_resources(
    ctx: BIContext, exp: Experiment, resources: list[str] | None, include_expired: bool, include_terminated: bool
) -> Collection[ComputeResource]:
    """Get requested resources for a bi command."""
    if not resources:  # all resources
        return ctx.obj.get_bi_resources_api_by_experiment(exp).list(
            include_expired=include_expired,
            include_terminated=include_terminated,
        )

    return [bi_get_resource(ctx, exp, resource_id) for resource_id in resources]


def handle_some_exceptions(func):
    """Handle specific exceptions by giving the user a useful error message instead of a confusing stacktrace."""

    @functools.wraps(func)
    def wrapper_handle_some_exceptions(*args, **kwargs):
        """Function wrapper for handle_some_exceptions."""  # noqa: D401
        try:
            return func(*args, **kwargs)
        except ApiResponseError as e:
            if e._httpx_status_error.response.status_code in (
                502,
                503,
                504,
            ):
                err_console.print("[red]There was a connection problem. The service might be down. :zap:")
                err_console.print(f"[red]Details: {e}")
                raise typer.Exit(code=1) from None
            # If it's something else, re-raise the exception and handle as usually
            raise
        except MaxRetryError as e:
            if isinstance(e.reason, NewConnectionError | HTTPSConnection):
                # example:
                #   NewConnectionError -> [Failed to establish a new connection: Errno 113] No route to host
                #   HTTPSConnection -> Failed to establish a new connection: [Errno 111] Connection refused
                err_console.print("[red]There was a connection problem. The service might be down. :zap:")
                err_console.print(f"[red]Details: {e.reason}")
                raise typer.Exit(code=1) from None
            # If it's something else, re-raise the exception and handle as usually
            raise

    return wrapper_handle_some_exceptions


def bi_get_experiment(ctx: BIContext, experiment_id: str) -> Experiment:
    """Get an experiment for a bi command."""
    if is_type_id(experiment_id, type="exp"):
        exp_obj = ctx.obj.api.experiment.get(experiment_id)
    else:
        exp_obj = ctx.obj.api.get_experiment(experiment_id, project_id=ctx.obj.require_project_id())

    if not exp_obj:
        err_console.print(f"Could not retrieve experiment {experiment_id} :x:")
        raise typer.Exit(code=1)
    return exp_obj


def _can_connect(host: str, port: int) -> bool:
    # TODO the timeout value is too short. Ideally, it should be based on known RTT of a slices service.

    try:
        with socket.socket(socket.AF_INET6, socket.SOCK_STREAM) as s:
            s.settimeout(0.1)
            if s.connect_ex((host, port)) == 0:
                # if we can connect, return True
                return True
    except:  # noqa: E722 S110
        # ignore IPv6 problems
        pass

    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.settimeout(0.1)
            return s.connect_ex((host, port)) == 0
    except:  # noqa: E722
        # Any exception means we can't connect.
        # Examples:
        #      gaierror: [Errno -5] No address associated with hostname
        return False


def bi_get_best_ssh_login(compute_resource: ComputeResource, with_proxy: bool) -> tuple[SSHInfo | None, bool]:
    """Get the best SSH login for a resource."""
    if not compute_resource.ssh_logins:
        return None, False

    best_ssh_login = None
    best_ssh_login_can_connect = False
    best_score = -1

    # Prefer the first one that has a public key
    for ssh_login in compute_resource.ssh_logins:
        score = 0

        # User wants to use a proxy
        if with_proxy and ssh_login.jump_proxy:
            score += 1000

        # Prefer connecting to a public IP
        if compute_resource.public_ipv4 and compute_resource.public_ipv4 in ssh_login.host:
            score += 100

        # Prefer connecting without needing to use the proxy
        if can_connect := _can_connect(ssh_login.host, ssh_login.port):
            score += 10

        if score > best_score:
            best_ssh_login = ssh_login
            best_ssh_login_can_connect = can_connect
            best_score = score

    # Otherwise, just return the first one
    return best_ssh_login, best_ssh_login_can_connect



"""BI CLI."""

import functools
from enum import Enum
from typing import Annotated

import typer
from msgspec.json import encode as json_encode
from msgspec.yaml import encode as yaml_encode
from rich.table import Table
from slices_clientlib_bi.bi_image import DiskImageCapability
from slices_clientlib_core.base.exceptions import ApiException, ServiceException
from urllib3.connection import HTTPSConnection
from urllib3.exceptions import MaxRetryError, NewConnectionError

from slices_cli_bi import console, err_console
from slices_cli_bi.util.errors import get_api_error_detail
from slices_cli_bi.util.time import friendly_datetime
from slices_cli_bi.util.type_id import get_type_id_domain

from .bi_context import BIContext

disk_images_app = typer.Typer(no_args_is_help=True)


def handle_some_exceptions(func):
    """Handle specific exceptions by giving the user a useful error message instead of a confusing stacktrace."""

    @functools.wraps(func)
    def wrapper_handle_some_exceptions(*args, **kwargs):
        """Function wrapper for handle_some_exceptions."""  # noqa: D401
        try:
            return func(*args, **kwargs)
        except ServiceException as e:
            err_console.print("[red]There was a connection problem. The service might be down. :zap:")
            err_console.print(f"[red]Details: {e}")
            raise typer.Exit(code=1) from e
        except MaxRetryError as e:
            if isinstance(e.reason, NewConnectionError | HTTPSConnection):
                # example:
                #   NewConnectionError -> [Failed to establish a new connection: Errno 113] No route to host
                #   HTTPSConnection -> Failed to establish a new connection: [Errno 111] Connection refused
                err_console.print("[red]There was a connection problem. The service might be down. :zap:")
                err_console.print(f"[red]Details: {e.reason}")
                raise typer.Exit(code=1) from None
            # If it's something else, re-raise the exception and handle as usually
            raise

    return wrapper_handle_some_exceptions


@disk_images_app.command(name="register", rich_help_panel="Admin Commands")
@handle_some_exceptions
def register_image(
    ctx: BIContext,
    friendly_name: Annotated[str, typer.Option(prompt=True, show_default=False)],
    location: Annotated[str, typer.Option(prompt=True, show_default=False)],
    default_username: Annotated[str, typer.Option(prompt=True, show_default=False)],
    capabilities: Annotated[
        list[DiskImageCapability],
        typer.Option("--capability", "--cap", help="Capabilities of the disk image."),
    ] = [  # noqa: B006 # type: ignore[reportArgumentType]
        DiskImageCapability.PING,
        DiskImageCapability.SSH,
        DiskImageCapability.CLOUD_MINUS_INIT,
        DiskImageCapability.CLOUD_MINUS_INIT_MINUS_PHONEHOME,
    ],
    short_description: Annotated[
        str | None, typer.Option(prompt=True, help="An optional short description of the image.")
    ] = None,
    long_description: Annotated[
        str | None, typer.Option(prompt=True, help="An optional long description of the image.")
    ] = None,
    min_disk: Annotated[
        int | None,
        typer.Option(show_default=False, help="The minimum size of the root disk, in MB."),
    ] = None,
    min_ram: Annotated[
        int | None,
        typer.Option(show_default=False, help="The minimum required amount of RAM for this image, MiB."),
    ] = None,
    tags: Annotated[list[str] | None, typer.Option(show_default=False)] = None,
):
    """Register disk images.

    Only available to admin users.
    """
    try:
        disk_image = (
            ctx.obj.api.for_project(ctx.obj.require_project_id())
            .bi(ctx.obj.infra_id)
            .image.create(
                friendly_name=friendly_name,
                cluster_id=ctx.obj.cluster_id,
                location=location,
                default_username=default_username,
                min_disk_mb=min_disk,
                min_ram_mib=min_ram,
                tags=tags,
                capabilities=capabilities,
                short_description=short_description,
                long_description=long_description,
            )
        )
        console.print(f"[bold green]:cd: Registered {friendly_name} as {disk_image.id} :white_check_mark:")
    except ApiException as err:
        err_console.print("[bold red]Could not register disk image[/] :x:")
        err_console.print(f"[red]{get_api_error_detail(err)}")
        raise typer.Exit(code=1) from None


class ListFormat(Enum):
    """List Formats."""

    table = "table"
    json = "json"
    yaml = "yaml"


def _extract_infa_id(any_type_id: str) -> str:
    return any_type_id.split("_")[1]


class ListImageSort(Enum):
    """List sort options."""

    name = "name"  # pyright: ignore [reportIncompatibleMethodOverride,reportAssignmentType] seems to work fine
    created = "created"


@disk_images_app.command(name="list", rich_help_panel="Commands")
@handle_some_exceptions
def list_images(
    ctx: BIContext,
    format: Annotated[
        ListFormat,
        typer.Option("--format", "-f", help="Format in which the images should be listed."),
    ] = "table",  # type: ignore
    show_tags: Annotated[
        bool,
        typer.Option(
            "--tags/--no-tags",
            "-t/-nt",
            help="Show/hide long description of image. (default: show)",
            rich_help_panel="Table view column selection",
            show_default=False,
        ),
    ] = True,
    show_short_description: Annotated[
        bool,
        typer.Option(
            "--desc/--no-desc",
            "-d/-nd",
            help="Show/hide short description of image. (default: show)",
            rich_help_panel="Table view column selection",
            show_default=False,
        ),
    ] = True,
    show_capabilities: Annotated[
        bool,
        typer.Option(
            "--capabilities/--no-capabilities",
            "-c/-nc",
            help="Show/hide capabilities. (default: hide)",
            rich_help_panel="Table view column selection",
            show_default=False,
        ),
    ] = False,
    show_long_description: Annotated[
        bool,
        typer.Option(
            "--long-desc/--no-long-desc",
            "-l/-nl",
            help="Show/hide long description of image. (default: hide)",
            rich_help_panel="Table view column selection",
            show_default=False,
        ),
    ] = False,
    show_id: Annotated[
        bool,
        typer.Option(
            "--id/--no-id",
            "-i/-ni",
            help="Show/hide the internal ID of the image. (default: hide)",
            rich_help_panel="Table view column selection",
            show_default=False,
        ),
    ] = False,
    show_infra: Annotated[
        bool,
        typer.Option(
            "--infra/--no-infra",
            "-I/-nI",
            help="Show/hide the infrastructure ID of the image. (default: show)",
            rich_help_panel="Table view column selection",
            show_default=False,
        ),
    ] = True,
    sort: Annotated[
        ListImageSort, typer.Option("--sort", "-s", help="Field to sort by (infra_id is always first sort field).")
    ] = "created",  # type: ignore
):
    """List disk images."""
    disk_images = ctx.obj.api.for_project(ctx.obj.require_project_id()).bi(ctx.obj.infra_id).image.list()

    disk_images = list(disk_images)  # fetch all diskimages now

    if sort == ListImageSort.name:
        disk_images.sort(key=lambda i: (get_type_id_domain(i.id), i.friendly_name, i.created_at))
    elif sort == ListImageSort.created:
        disk_images.sort(key=lambda i: (get_type_id_domain(i.id), i.created_at, i.friendly_name))

    if format == ListFormat.table:
        cols = ["Name"]
        if show_infra:
            cols.insert(0, "Infra ID")
        if show_id:
            cols.insert(0, "ID")
        if show_capabilities:
            cols.append("Capabilities")
        cols.append("Created At")
        if show_tags:
            cols.append("Tags")
        if show_short_description:
            cols.append("Description")
        if show_long_description:
            cols.append("Detailed Description")

        table = Table(*cols, title="Images")

        for disk_image in disk_images:
            row = [
                disk_image.friendly_name,
            ]
            if show_infra:
                row.insert(0, _extract_infa_id(disk_image.id))
            if show_id:
                row.insert(0, disk_image.id)
            if show_capabilities:
                row.append(", ".join(sorted(disk_image.capabilities or [])) or "-")

            row.append(
                friendly_datetime(disk_image.created_at) if disk_image.created_at else "-",
            )

            if show_tags:
                row.append(", ".join(disk_image.tags or []) or "-")
            if show_short_description:
                row.append(disk_image.short_description or "-")
            if show_long_description:
                row.append(disk_image.long_description or "-")

            table.add_row(*row)

        console.print(table)

    elif format == ListFormat.json:
        console.print_json(json_encode(list(disk_images)).decode())
    elif format == ListFormat.yaml:
        console.print(yaml_encode(list(disk_images)).decode())



"""Error handling utilities."""

import json
from json import JSONDecodeError
from typing import Any

from slices_clientlib_core.base.exceptions import (
    ApiException,
    BadRequestException,
    NotFoundException,
    ServiceException,
    UnauthorizedException,
    ValidationError,
)


def get_error_detail(error_body: Any | None):
    """Get detail from error response."""
    if not isinstance(error_body, str):
        return None
    try:
        error_body_contents = json.loads(error_body)
    except JSONDecodeError:
        return error_body

    error_details = []

    if "detail" in error_body_contents:
        if isinstance(error_body_contents["detail"], list):
            for detail in error_body_contents["detail"]:
                if "msg" in detail:
                    error_details.append(detail["msg"])
                else:
                    error_details.append(str(detail))
        else:
            error_details.append(error_body_contents["detail"])

    if error_details:
        res = []
        for detail in error_details:
            if isinstance(detail, str):
                res.append(detail)
            elif isinstance(detail, dict | list):
                res.append(json.dumps(detail, indent=4))
            else:
                res.append(str(detail))
        return "\n".join(res)

    return "No details available"


def get_api_error_detail(err: ApiException):
    """Show error details for API exceptions."""

    if isinstance(err, ValidationError):
        return "\n".join([f"{error_detail.loc}: {error_detail.msg}" for error_detail in err.details])

    if isinstance(err, BadRequestException):
        return f"Bad Request: {err.detail}"

    if isinstance(err, ServiceException):
        message = "The server encountered an internal error"
        if err._httpx_error:
            message += (
                f" while processing {err._httpx_error.request.method} request "
                f"on {err._httpx_error.request.url}: {err._httpx_status_error.response.text}"
            )
        else:
            message += "."
        message += "\n\nPlease report this issue to support if the issue persists."

        return message

    if isinstance(err, UnauthorizedException):
        return f"Unauthorized: {err._httpx_error}"

    if isinstance(err, NotFoundException):
        assert err._httpx_error
        return f"Resource not found: {err._httpx_error.request.url}"

    return f"TODO: Implement get_api_error_detail for other error types {err} {err!r}"



"""BI Flavor CLI."""

import functools
from enum import Enum
from typing import Annotated

import typer
from msgspec.json import encode as json_encode
from msgspec.yaml import encode as yaml_encode
from rich.table import Table
from slices_clientlib_bi.bi_flavor import AnyFlavor, BareMetalFlavor, BareMetalFlavorCreate, VMFlavor, VMFlavorCreate
from slices_clientlib_core.base.exceptions import (
    ApiException,
    BadRequestException,
    ForbiddenException,
    ServiceException,
)
from urllib3.connection import HTTPSConnection
from urllib3.exceptions import MaxRetryError, NewConnectionError

from slices_cli_bi import console, err_console
from slices_cli_bi.util.errors import get_api_error_detail
from slices_cli_bi.util.type_id import get_type_id_domain

from .bi_context import BIContext

flavors_app = typer.Typer(no_args_is_help=True)


def handle_some_exceptions(func):
    """Handle specific exceptions by giving the user a useful error message instead of a confusing stacktrace."""

    @functools.wraps(func)
    def wrapper_handle_some_exceptions(*args, **kwargs):
        """Function wrapper for handle_some_exceptions."""  # noqa: D401
        try:
            return func(*args, **kwargs)
        except ServiceException as e:
            err_console.print("[red]There was a connection problem. The service might be down. :zap:")
            err_console.print(f"[red]Details: {get_api_error_detail(e)}")
            raise typer.Exit(code=1) from None
        except BadRequestException as e:
            if "Project expired".lower() in e._httpx_status_error.response.text.lower():
                # Typically this occurs on experiment create
                err_console.print("[red]The project has expired. :x:")
                raise typer.Exit(code=1) from None
            # If it's something else, re-raise the exception and handle as usually
            raise
        except MaxRetryError as e:
            if isinstance(e.reason, NewConnectionError | HTTPSConnection):
                # example:
                #   NewConnectionError -> [Failed to establish a new connection: Errno 113] No route to host
                #   HTTPSConnection -> Failed to establish a new connection: [Errno 111] Connection refused
                err_console.print("[red]There was a connection problem. The service might be down. :zap:")
                err_console.print(f"[red]Details: {e.reason}")
                raise typer.Exit(code=1) from None
            # If it's something else, re-raise the exception and handle as usually
            raise

    return wrapper_handle_some_exceptions


# Not supported at the moment
# @flavors_app.command(name="create", rich_help_panel="Commands")
# def create_flavor(
#     ctx: BIContext,
#     friendly_name: Annotated[
#         str,
#         typer.Option(
#             prompt=True,
#             show_default=False,
#             help='Short user readable name. On the current BIs, this should start with "custom."',
#         ),
#     ],
#     description: Annotated[str, typer.Option(prompt=True, show_default=False, help="Description of this flavor")],
#     ram: Annotated[int, typer.Option(prompt=True, show_default=False, help="The amount of RAM a flavor has, in MiB")],
#     vcpus: Annotated[
#         int,
#         typer.Option(
#             prompt=True, show_default=False, help="The number of virtual CPUs that will be allocated to the server"
#         ),
#     ],
#     disk: Annotated[
#         int,
#         typer.Option(prompt=True, show_default=False, help="The size of the root disk that will be created in MB."),
#     ],
# ):
#     """Create a flavor.
#
#     This flavor is stored in the project.
#     """
#     try:
#         flavor = (
#             ctx.obj.api.for_project(ctx.obj.require_project_id())
#             .bi(ctx.obj.infra_id)
#             .flavor.create_in_project(
#                 VMFlavorCreate(
#                     friendly_name=friendly_name,
#                     description=description,
#                     ram_mib=ram,
#                     vcpus=vcpus,
#                     disk_mb=disk,
#                 )
#             )
#         )
#         console.print(f"[bold green]:computer: Created {friendly_name} as {flavor.id} :white_check_mark:")
#     except BadRequestException as exc:
#         err_console.print("[bold red]Could not create flavor[/] :x:")
#         err_console.print(f"[red] {get_api_error_detail(exc)}")
#         raise typer.Exit(code=1) from None
#     except ForbiddenException:
#         err_console.print("[bold red]You do not have permission to create a flavor.[/] :x:")
#         raise typer.Exit(code=1) from None
#     except ApiException as exc:
#         err_console.print("[bold red]Could not create flavor[/] :x:")
#         err_console.print(f"[red] {get_api_error_detail(exc)}")
#         raise typer.Exit(code=1) from None


class FlavorType(str, Enum):
    """The type of flavor."""

    vm = "vm"
    baremetal = "baremetal"


@flavors_app.command(name="create-global", rich_help_panel="Admin Commands")
def create_global_flavor(
    ctx: BIContext,
    friendly_name: Annotated[
        str,
        typer.Option(
            prompt=True,
            show_default=False,
            help='Short user readable name. On the current BIs, this should start with "custom."',
        ),
    ],
    description: Annotated[str, typer.Option(prompt=True, show_default=False, help="Description of this flavor")],
    type: Annotated[FlavorType, typer.Option(case_sensitive=True, help="The type of flavor.")],
    ram: Annotated[int, typer.Option(prompt=True, show_default=False, help="The amount of RAM a flavor has, in MiB")],
    disk: Annotated[
        int,
        typer.Option(prompt=True, show_default=False, help="The size of the root disk that will be created in MB."),
    ],
    vcpus: Annotated[
        int,
        typer.Option(
            prompt=True, show_default=False, help="The number of virtual CPUs that will be allocated to the server"
        ),
    ]
    | None = None,
    cpu_model: Annotated[
        str,
        typer.Option(prompt=True, show_default=False, help="The CPU model"),
    ]
    | None = None,
    cpu_sockets: Annotated[
        int,
        typer.Option(prompt=True, show_default=False, help="The number of Physical CPU sockets"),
    ]
    | None = None,
    cpu_cores: Annotated[
        int,
        typer.Option(prompt=True, show_default=False, help="The number of Physical CPU cores (per CPU socket)"),
    ]
    | None = None,
    cpu_threads: Annotated[
        int,
        typer.Option(prompt=True, show_default=False, help="The number of Physical CPU Threads (per CPU core)"),
    ]
    | None = None,
    gpu_model: Annotated[
        str,
        typer.Option(prompt=True, show_default=False, help="The GPU model (if a GPU is present)"),
    ]
    | None = None,
    network_interfaces: Annotated[
        int,
        typer.Option(prompt=True, show_default=False, help="The number of network interfaces"),
    ]
    | None = None,
):
    """Create a global flavor, available for everyone. Admin only."""
    try:
        if type == FlavorType.baremetal:
            if cpu_cores is None:
                err_console.print("[red] --cpu-cores required")
                raise typer.Exit(code=1) from None
            if cpu_model is None:
                err_console.print("[red] --cpu-model required")
                raise typer.Exit(code=1) from None
            if cpu_sockets is None:
                err_console.print("[red] --cpu-sockets required")
                raise typer.Exit(code=1) from None
            if cpu_threads is None:
                err_console.print("[red] --cpu-threads required")
                raise typer.Exit(code=1) from None
            flavor_create = BareMetalFlavorCreate(
                friendly_name=friendly_name,
                description=description,
                ram_mib=ram,
                disk_mb=disk,
                cpu_cores=cpu_cores,
                cpu_model=cpu_model,
                cpu_sockets=cpu_sockets,
                cpu_threads=cpu_threads,
                gpu_model=gpu_model,
                network_interface_count=network_interfaces or 0,
            )
        elif type == FlavorType.vm:
            if vcpus is None:
                err_console.print("[red] --vcpus required")
                raise typer.Exit(code=1) from None
            flavor_create = VMFlavorCreate(
                friendly_name=friendly_name,
                description=description,
                ram_mib=ram,
                vcpus=vcpus,
                disk_mb=disk,
            )
        else:
            raise ValueError(f"Unsupported flavor type {type}")

        flavor = (
            ctx.obj.api.for_project(ctx.obj.require_project_id())
            .bi(ctx.obj.infra_id)
            .flavor.create_global(flavor_create)
        )
        console.print(f"[bold green]:computer: Created {friendly_name} as {flavor.id} :white_check_mark:")
    except BadRequestException as exc:
        err_console.print("[bold red]Could not create flavor[/] :x:")
        err_console.print(f"[red] {get_api_error_detail(exc)}")
        raise typer.Exit(code=1) from None
    except ForbiddenException:
        err_console.print("[bold red]You do not have permission to create a flavor.[/] :x:")
        raise typer.Exit(code=1) from None
    except ApiException as exc:
        err_console.print("[bold red]Could not create flavor[/] :x:")
        err_console.print(f"[red] {get_api_error_detail(exc)}")
        raise typer.Exit(code=1) from None


def _float_to_str(v: float) -> str:
    if float(round(v)) == v:
        return str(int(v))
    return str(round(v, 3))


def _extract_infa_id(any_type_id: str) -> str:
    return any_type_id.split("_")[1]


def _print_vm_flavors_table(infra: str, flavors: list, show_infra: bool, show_id: bool):
    names = ["Name", "Desc", "(v)CPUs", "RAM (GiB)", "Root Disk (GB)"]

    if show_infra:
        names.insert(0, "Infra")
    if show_id:
        names.insert(0, "ID")

    table = Table(
        *names,
        title=f"Slices BI Flavors on {infra}" if infra != "bi-orchestrator" else "Slices BI VM Flavors",
    )
    for flavor in flavors:
        if isinstance(flavor, VMFlavor):
            values = [
                flavor.friendly_name,
                flavor.description,
                str(flavor.vcpus),
                _float_to_str(flavor.ram_mib / 1024.0),
                _float_to_str(flavor.disk_mb / 1000.0),
            ]
            if show_infra:
                values.insert(0, _extract_infa_id(flavor.id))
            if show_id:
                values.insert(0, flavor.id)
            table.add_row(*values)

    console.print(table)


def _print_bare_metal_flavors_table(infra: str, flavors: list, show_infra: bool, show_id: bool):
    names = ["Name", "Desc", "CPUs", "RAM (GiB)", "Root Disk (GB)", "Interfaces", "GPU"]

    if show_infra:
        names.insert(0, "Infra")
    if show_id:
        names.insert(0, "ID")

    table = Table(
        *names,
        title=f"Slices BI Flavors on {infra}" if infra != "bi-orchestrator" else "Slices BI Baremetal Flavors",
    )
    for flavor in flavors:
        if isinstance(flavor, BareMetalFlavor):
            values = [
                flavor.friendly_name,
                flavor.description,
                f"{flavor.cpu_sockets}x{flavor.cpu_cores}x{flavor.cpu_threads}",
                _float_to_str(flavor.ram_mib / 1024.0),
                _float_to_str(flavor.disk_mb / 1000.0),
                str(flavor.network_interface_count),
                flavor.gpu_model if flavor.gpu_model else "-",
            ]
            if show_infra:
                values.insert(0, _extract_infa_id(flavor.id))
            if show_id:
                values.insert(0, flavor.id)
            table.add_row(*values)
    console.print(table)


class ListFormat(Enum):
    """List Formats."""

    table = "table"
    json = "json"
    yaml = "yaml"


def _flavor_sort_helper(flavor: AnyFlavor):
    if isinstance(flavor, BareMetalFlavor):
        return f"bm_{flavor.friendly_name}"
    if isinstance(flavor, VMFlavor):
        return f"vm_{flavor.vcpus:03d}_{flavor.ram_mib:07d}_{flavor.disk_mb:09d}_{flavor.friendly_name}"
    return f"xx_{flavor.friendly_name}"


@flavors_app.command(name="list", rich_help_panel="Commands")
@handle_some_exceptions
def list_flavors(
    ctx: BIContext,
    show_id: Annotated[
        bool,
        typer.Option(
            "--id/--no-id",
            "-i/-ni",
            help="Show/hide the internal ID of the flavor. (default: hide)",
            rich_help_panel="Table view column selection",
            show_default=False,
        ),
    ] = False,
    show_infra: Annotated[
        bool,
        typer.Option(
            "--infra/--no-infra",
            "-I/-nI",
            help="Show/hide the infrastructure ID of the flavor. (default: show)",
            rich_help_panel="Table view column selection",
            show_default=False,
        ),
    ] = True,
    format: Annotated[
        ListFormat,
        typer.Option("--format", "-f", help="Format in which the flavors should be listed."),
    ] = "table",  # type: ignore
):
    """List flavors."""
    flavors = ctx.obj.api.for_project(ctx.obj.require_project_id()).bi(ctx.obj.infra_id).flavor.list()

    # we need to run over the flavors twice, so we need to store them in a list
    flavors = list(flavors)
    flavors.sort(key=lambda f: (get_type_id_domain(f.id), _flavor_sort_helper(f)))

    if format == ListFormat.table:
        has_bare_metal = False
        has_vm = False

        for flavor in flavors:
            if isinstance(flavor, BareMetalFlavor):
                has_bare_metal = True
            elif isinstance(flavor, VMFlavor):
                has_vm = True

        infra = ctx.obj.infra_id

        if has_vm:
            _print_vm_flavors_table(flavors=flavors, infra=infra, show_infra=show_infra, show_id=show_id)
        if has_bare_metal:
            _print_bare_metal_flavors_table(flavors=flavors, infra=infra, show_infra=show_infra, show_id=show_id)

        # For reason of difference between MB and MiB for disk/ram, see https://en.wikipedia.org/wiki/Gigabyte
        # M/Mi instead of G/Gi everywhere to be flexible for IoT and other small devices

    elif format == ListFormat.json:
        console.print_json(json_encode(list(flavors)).decode())
    elif format == ListFormat.yaml:
        console.print(yaml_encode(list(flavors)).decode())



"""CLI configuration management."""

import logging
import os
from dataclasses import dataclass
from functools import cache
from typing import TypeVar

from slices_clientlib_core.base.config import config_dir_location
from slices_clientlib_core.base.exceptions import ConfigurationError
from tomlkit import TOMLDocument, comment, nl, table
from tomlkit.exceptions import TOMLKitError
from tomlkit.toml_file import TOMLFile

logger = logging.getLogger(__name__)


@dataclass
class BISettings:
    """SSH settings for BI CLI."""

    extra_ssh_options: list[str]
    extra_sftp_options: list[str]
    extra_scp_options: list[str]
    ssh_exe: str
    sftp_exe: str
    scp_exe: str


bi_settings_file = config_dir_location / "bi_settings.toml"


def _get_bi_ssh_settings_toml():
    try:
        # If there is no SSH settings file, write it.
        # This file with default settings serves as a template for users to customize.
        if not bi_settings_file.exists():
            write_default_bi_settings()
    except TOMLKitError:
        # ignore not being able to write this file
        pass

    try:
        if bi_settings_file.exists():
            settings_file = TOMLFile(bi_settings_file)
            return settings_file.read()
    except TOMLKitError as e:
        raise ConfigurationError(f"Could not read settings file {bi_settings_file}") from e

    return TOMLDocument()


T = TypeVar("T")


def _env_or_toml(env_key: str, toml_section: str, toml_key: str, toml: TOMLDocument, default: T) -> T:
    """Get a value from the environment or TOML file."""
    if env_key in os.environ:
        return os.environ[env_key] if not isinstance(default, list) else os.environ[env_key].split()  # pyright: ignore[reportReturnType]
    if toml_section in toml and toml_key in toml[toml_section]:  # pyright: ignore[reportOperatorIssue]
        return toml[toml_section][toml_key]  # pyright: ignore[reportIndexIssue,reportReturnType]
    return default


@cache
def get_bi_settings() -> BISettings:
    """Get BI SSH settings."""
    toml = _get_bi_ssh_settings_toml()
    return BISettings(
        extra_ssh_options=_env_or_toml("SLICES_BI_EXTRA_SSH_OPTIONS", "ssh", "extra_options", toml, []),
        extra_scp_options=_env_or_toml("SLICES_BI_EXTRA_SCP_OPTIONS", "scp", "extra_options", toml, []),
        extra_sftp_options=_env_or_toml("SLICES_BI_EXTRA_SFTP_OPTIONS", "sftp", "extra_options", toml, []),
        ssh_exe=_env_or_toml("SLICES_BI_SSH_EXE", "ssh", "exe", toml, "ssh"),
        scp_exe=_env_or_toml("SLICES_BI_SCP_EXE", "scp", "exe", toml, "scp"),
        sftp_exe=_env_or_toml("SLICES_BI_SFTP_EXE", "sftp", "exe", toml, "sftp"),
    )


def write_default_bi_settings():
    """Write the settings file with default BI SSH settings."""
    toml = TOMLDocument()
    toml.add(comment("This file can be used to customize the Slices BI CLI."))
    # toml.add(nl())
    toml.add(comment("Each of the settings can be overridden with a specific SLICES_BI_* environment variables."))
    toml.add(nl())

    ssh_section = table()
    ssh_section.add(comment("These options control the command: slices bi ssh"))
    ssh_section.add(comment(""))
    ssh_section.add(comment("extra_options:"))
    ssh_section.add(comment("    Extra options to pass to the ssh command, before any other options."))
    ssh_section.add(comment("    Env var: SLICES_BI_EXTRA_SSH_OPTIONS"))
    ssh_section.add(comment(""))
    toml.add("ssh", ssh_section)
    ssh_section["extra_options"] = []
    ssh_section.add(comment("Example:"))
    ssh_section.add(comment('extra_options = ["-A", "-X", "-oPasswordAuthentication=no"]'))
    # ssh_section["exe"] = "ssh"  # not documented at the moment
    # ssh_section["exe"].comment("SLICES_BI_SFTP_EXE")  # not documented at the moment

    scp_section = table()
    scp_section.add(comment("These options control the command: slices bi scp"))
    scp_section.add(comment(""))
    scp_section.add(comment("extra_options:"))
    scp_section.add(comment("    Extra options to pass to the scp command, before any other options."))
    scp_section.add(comment("    Env var: SLICES_BI_EXTRA_SCP_OPTIONS"))
    scp_section.add(comment(""))
    toml.add("scp", scp_section)
    scp_section["extra_options"] = []
    # scp_section["exe"] = "scp"  # not documented at the moment
    # scp_section["exe"].comment("SLICES_BI_SCP_EXE")  # not documented at the moment

    sftp_section = table()
    sftp_section.add(comment("These options control the command: slices bi sftp"))
    sftp_section.add(comment(""))
    sftp_section.add(comment("extra_options:"))
    sftp_section.add(comment("    Extra options to pass to the sftp command, before any other options."))
    sftp_section.add(comment("    Env var: SLICES_BI_EXTRA_SFTP_OPTIONS"))
    sftp_section.add(comment(""))
    toml.add("sftp", sftp_section)
    sftp_section["extra_options"] = []
    # sftp_section["exe"] = "sftp"  # not documented at the moment
    # sftp_section["exe"].comment("SLICES_BI_SFTP_EXE")  # not documented at the moment

    config_dir_location.mkdir(exist_ok=True)
    settings_file = TOMLFile(bi_settings_file)
    settings_file.write(toml)



"""Store extra error details."""

import sys
from contextlib import AbstractContextManager
from datetime import datetime
from io import StringIO
from pathlib import Path

from msgspec import Struct
from msgspec._core import json_encode, json_format
from pytz import UTC
from rich.console import Console
from rich.traceback import Trace, Traceback

from slices_cli_bi import console, err_console
from slices_cli_bi.core.const import CONFIG_DIR
from slices_cli_bi.util.shell import shlex_quote_if_needed


class ErrorDetailsHandler(AbstractContextManager):
    """A context manager for storing extra error details.

    ErrorDetailsHandler is typically created when an error occurs.
    The contextmanager exits and writes an error details file when the Exit is raised.
    It also outputs a message to the user, showing where the file is written.
    """

    def __init__(self):
        self.console_io = StringIO()
        self.console = Console(
            file=self.console_io,
            force_terminal=False,
            force_jupyter=False,
            force_interactive=False,
            width=140,
            height=30,
            no_color=True,
        )
        self._objects: list[tuple[str, Struct]] = []
        self._traces: list[Trace] = []
        # You can make an ErrorDetailsHandler before you now there's an error,
        # and only set this to True when the error occurs
        # This allows older context to be logged.
        self.must_store_failure = True

    def __exit__(self, *exc_details):
        """When exiting the ErrorDetailsHandler context manager, write details to file."""
        if self.must_store_failure:
            self._store_failure()

    def print(self, *args, **kwargs):
        """Print to stderr console."""
        err_console.print(*args, **kwargs)
        self.console.print(*args, **kwargs)

    def print_stdout(self, *args, **kwargs):
        """Print to stdout console."""
        console.print(*args, **kwargs)
        self.console.print(*args, **kwargs)

    def add_object(self, name: str, obj: Struct):
        """Add an object to the Error Details."""
        self._objects.append((name, obj))

    def add_cur_trace(self):
        """Add the current trace to the Error Details."""
        exc_type, exc_value, traceback = sys.exc_info()
        if exc_type is None or exc_value is None or traceback is None:
            self.console.print("ErrorDetailsHandler.add_cur_trace should be called within except block!")
            return
        self._traces.append(Traceback.extract(exc_type, exc_value, traceback, show_locals=True))

    def _store_failure(self) -> Path:
        error_details_dir = Path.home() / CONFIG_DIR / "cli_error_details"
        if not error_details_dir.exists():
            error_details_dir.mkdir()
        now = datetime.now(UTC)
        err_file = error_details_dir / f"err_{now.isoformat(sep='T', timespec='seconds')}.txt"
        with err_file.open("w") as f:
            file_console = Console(
                file=f,
                force_terminal=False,
                force_jupyter=False,
                force_interactive=False,
                width=140,
                height=30,
                no_color=True,
            )

            def print_head(txt: str):
                count = 4 + len(txt)
                file_console.print("#" * count)
                file_console.print(f"# {txt} #")
                file_console.print("#" * count)

            print_head("Error details.")
            command = " ".join([shlex_quote_if_needed(a) for a in sys.argv])
            file_console.print(f"Command: {command}")
            file_console.print(f"Date: {datetime.now(UTC)}")
            file_console.print()

            print_head("Console Error Output")
            file_console.print(self.console_io.getvalue())
            file_console.print()

            for name, obj in self._objects:
                print_head(f"Object {name!r}")
                file_console.print(json_format(json_encode(obj).decode()))
                file_console.print()

            for index, trace in enumerate(self._traces):
                print_head(f"Traceback {index}")
                file_console.print(Traceback(trace))
                file_console.print()

        console.print(f"[bright_black]Details stored at: {str(err_file.absolute())}", highlight=False)
        return err_file



"""Utility to follow progress of a BI task."""

import os
from time import sleep

from rich.progress import BarColumn, Progress, TaskID, TaskProgressColumn, TextColumn
from slices_clientlib_bi.bi_task import BiTaskClient, TaskStatus

from slices_cli_bi import console

MAX_PROGRESS_UPDATES_PER_REQUEST = 40  # stay within the limit of how long a GET-url can become

progress_columns = [
    TextColumn("[progress.description]{task.description}"),
    BarColumn(),
    TaskProgressColumn(),
]


class SimpleProgressPrint:
    """A very simple status printer.

    It keeps overwriting the same line, by using carriage return.
    """

    SIMPLE_PROGRESS_CHARS = ["/", "-", "\\", "|"]

    def __init__(self, single_task: bool):
        try:
            self.simple_progress_cols = int(os.getenv("COLUMNS", "30"))
        except ValueError:
            self.simple_progress_cols = 30
        self.char_index = 0
        self.single_task = single_task

    def print(self, status: TaskStatus | None):
        """Print a status line."""
        ch = self.SIMPLE_PROGRESS_CHARS[self.char_index % len(self.SIMPLE_PROGRESS_CHARS)]
        if self.single_task and status:
            perc = (status.completed or 0) * 100.0 / (status.total or status.completed or 100)
            console.print(f"{f'{perc}% {ch} {status.description}': <{self.simple_progress_cols}}", end="\r")
        else:
            console.print(f"{ch: <{self.simple_progress_cols}}", end="\r")
        self.char_index += 1

    def clear(self):
        """Clear the line."""
        console.print(" " * self.simple_progress_cols, end="\r")


def follow_progress(
    bi_tasks_api: BiTaskClient,
    task_ids: list[str],
    task_description_prefixes: list[str],
    *,
    transient=False,
) -> dict[str, TaskStatus]:
    """Periodically query BI tasks API to check on task while showing progress bar."""
    assert len(task_ids) == len(task_description_prefixes)
    unique_task_id_len = len(set(task_ids))

    # If the console is not a terminal, or a dumb terminal, progress won't do anything until it is done.
    if console.is_dumb_terminal or not console.is_terminal:
        simple_progress = SimpleProgressPrint(unique_task_id_len == 1)
    else:
        simple_progress = None

    finished_tasks: dict[str, TaskStatus] = {}
    a_task: TaskStatus | None = None
    with Progress(*progress_columns, transient=transient, console=console) as progress:
        bi_task_ids_to_progress_t_ids: dict[str, list[TaskID]] = {}
        bi_task_ids_to_description_prefixes: dict[str, list[str]] = {}

        for task_id, task_description_prefix in zip(task_ids, task_description_prefixes, strict=True):
            progress_t_id = progress.add_task(task_description_prefix, total=None)
            bi_task_ids_to_progress_t_ids.setdefault(task_id, []).append(progress_t_id)
            bi_task_ids_to_description_prefixes.setdefault(task_id, []).append(task_description_prefix)

        while not progress.finished:
            assert len(finished_tasks) < unique_task_id_len

            tasks_to_query = list({task_id for task_id in task_ids if task_id not in finished_tasks})
            bi_task_status: list[TaskStatus] = []
            while tasks_to_query:
                bi_task_status.extend(
                    bi_tasks_api.get_tasks(
                        task_ids=tasks_to_query[0:MAX_PROGRESS_UPDATES_PER_REQUEST],  # type: ignore
                    )
                )
                tasks_to_query = tasks_to_query[MAX_PROGRESS_UPDATES_PER_REQUEST:]

            for task_status in bi_task_status:
                a_task = task_status
                task_finished = task_status.status in ("finished", "failed")
                task_success = task_status.status == "finished"
                progress_t_ids = bi_task_ids_to_progress_t_ids[task_status.id]
                for progress_t_id in progress_t_ids:
                    completed = task_status.completed or 0
                    total = task_status.total or 100

                    # The API only use task.status for finished. task.completed and task.total are only informative.
                    # But rich.progress uses completed >= total to detect finished.
                    #
                    # So to use rich.progress we need to auto fix bad completed counts coming from the API:
                    # - If status=finished but `completed` is not `total`, set `completed` to `total`
                    #   This prevents finished from never being reached because backend code forgot to set 100%.
                    # - If not status=finished but `completed` equal to `total`, set `completed` to `total-1`
                    #   This one prevents the informative 100% from being considered as finished
                    #   before the task is really finished.
                    completed = min(completed, total - 1) if not task_finished else task_status.total

                    progress.update(progress_t_id, total=task_status.total, completed=completed)

                    task_description = task_status.description
                    if task_status.failure_reason:
                        # make the failure reason message show up here as well
                        if task_description:
                            task_description = task_description + " # " + task_status.failure_reason.message
                        else:
                            task_description = task_status.failure_reason.message
                    if task_description:
                        for description_prefix in bi_task_ids_to_description_prefixes[task_status.id]:
                            progress.update(
                                progress_t_id,
                                description=f"{description_prefix}: {task_description}",
                                visible=not task_success,
                            )

                if task_finished:
                    finished_tasks[task_status.id] = task_status

            if not progress.finished:
                if simple_progress:
                    simple_progress.print(a_task)
                sleep(0.5)
            elif simple_progress:
                # cleanup simple busy view
                simple_progress.clear()

    return finished_tasks


def follow_one_progress(
    bi_tasks_api: BiTaskClient,
    task_id,
    task_description_prefix,
    *,
    transient=False,
):
    """Periodically query BI tasks API to check on task while showing progress bar."""
    follow_progress(bi_tasks_api, [task_id], [task_description_prefix], transient=transient)
